from gutter.client.operators import Base
from gutter.client.registry import operators


class Truthy(Base):

    name = 'true'
    group = 'identity'
    preposition = 'true'

    def applies_to(self, argument):
        return bool(argument)

    def __str__(self):
        return 'true'


operators.register(Truthy)



import logging

from google.cloud import datastore
import pickle
import codecs

LOG_TOKEN = '[DATASTORE_LIKE_DICT] {}'


def encode(obj):
    return codecs.encode(pickle.dumps(obj), 'base64').decode()


def decode(string):
    return pickle.loads(codecs.decode(string.encode(), "base64"))


class DatastoreDict(dict):

    def __init__(self, namespace=None, kind=None):
        self.client = datastore.Client()
        self.kind = kind
        self.namespace = namespace
        logging.info(LOG_TOKEN.format(
            'Init :: Namespace {} :: Kind {}'.format(
                namespace,
                kind
            )
        ))

    def __setitem__(self, key, item):
        logging.info(LOG_TOKEN.format(
            'Insert :: key {} :: value {}'.format(
                key,
                item
            )
        ))

        _key = self.client.key(self.kind, key, namespace=self.namespace)
        _item = {
            'value': encode(item)
        }

        _entity = datastore.Entity(key=_key)
        _entity.update(_item)

        self.client.put(entity=_entity)

        # self.__dict__[key] = item

    def __getitem__(self, key):
        logging.info(LOG_TOKEN.format(
            'Get Item :: key {}'.format(
                key
            )
        ))

        _key = self.client.key(self.kind, key, namespace=self.namespace)
        _entity = self.client.get(_key)

        if _entity:
            return decode(_entity['value'])

        raise KeyError('Key not found')

    def __delitem__(self, key):
        logging.info(LOG_TOKEN.format(
            'Delete :: key {}'.format(
                key
            )
        ))

        _key = self.client.key(self.kind, key, namespace=self.namespace)
        self.client.delete(_key)

    def keys(self):
        logging.info(LOG_TOKEN.format(
            'Keys'
        ))

        _query = self.client.query(kind=self.kind, namespace=self.namespace)
        _result = list(_query.fetch())
        _return = dict()

        for _r in _result:
            _return[_r.key.name] = decode(_r['value'])

        return _return.keys()

    def items(self):
        logging.info(LOG_TOKEN.format(
            'Itens'
        ))

        _query = self.client.query(kind=self.kind, namespace=self.namespace)
        _result = list(_query.fetch())
        _return = dict()

        for _r in _result:
            _return[_r.key.name] = decode(_r['value'])

        return _return.items()

    def __contains__(self, item):
        logging.info(LOG_TOKEN.format(
            'Contains :: Key {}'.format(
                item
            )
        ))

        _key = self.client.key(self.kind, item, namespace=self.namespace)
        _entity = self.client.get(_key)

        if _entity:
            return True

        raise False



"""
gutter.models
~~~~~~~~~~~~~~~

:copyright: (c) 2010-2012 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from __future__ import absolute_import

# Standard Library
import threading
from collections import defaultdict
from functools import partial

# External Libraries
import six
from gutter.client import signals
from gutter.client.compat import ifilter

DEFAULT_SEPARATOR = ':'


def all_false_if_empty(iterable):
    if not iterable:
        return False

    for element in iterable:
        if not element:
            return False
    return True


class ConditionsDict(defaultdict):

    @classmethod
    def from_conditions_list(cls, conditions):
        conditions_dict = cls(set)

        for cond in conditions:
            conditions_dict[cond.argument.COMPATIBLE_TYPE].add(cond)

        return conditions_dict

    def get_by_input(self, inpt):
        return self.get_by_type(type(inpt))

    def get_by_type(self, to_get_key):
        if to_get_key in self:
            return self[to_get_key]
        for key_type in self:
            if issubclass(to_get_key, key_type):
                return self[key_type]

        # raise the correct exception
        return self[to_get_key]


class Switch(object):

    """
    A switch encapsulates the concept of an item that is either 'on' or 'off'
    depending on the input.  The switch determines this by checking each of its
    conditions and seeing if it applies to a certain input.  All the switch does
    is ask each of its Conditions if it applies to the provided input.  Normally
    any condition can be true for the Switch to be enabled for a particular
    input, but of ``switch.compounded`` is set to True, then **all** of the
    switches conditions need to be true in order to be enabled.

    See the Condition class for more information on what a Condition is and how
    it checks to see if it's satisfied by an input.
    """

    class states:
        DISABLED = 1
        SELECTIVE = 2
        GLOBAL = 3

    def __init__(
            self,
            name,
            state=states.DISABLED,
            compounded=False,
            concent=True,
            manager=None,
            label=None,
            description=None,
            **kwargs
    ):
        self._name = str(name)
        self.label = label
        self.description = description
        self.state = state
        self.conditions = []
        self.compounded = compounded
        self.concent = concent
        self.manager = manager
        self.reset()

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        separator = getattr(self.manager, 'key_separator', DEFAULT_SEPARATOR)
        parent = self.name.rsplit(separator, 1)[0]
        return parent if parent != self.name else None

    def get_parent(self):
        return self.manager.switch(self.parent) if self.parent else None

    def __repr__(self):
        kwargs = dict(
            state=self.state,
            compounded=self.compounded,
            concent=self.concent
        )
        parts = ["%s=%s" % (k, v) for k, v in kwargs.items()]
        return '<Switch("%s") conditions=%s, %s>' % (
            self.name,
            len(self.conditions),
            ', '.join(parts)
        )

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.state is other.state and
            self.compounded is other.compounded and
            self.concent is other.concent
        )

    def __getstate__(self):
        inner_dict = vars(self).copy()
        inner_dict.pop('manager', False)
        return inner_dict

    def __setstate__(self, state):
        # remove parent from the state, this is now a calculated field
        for attr in (
            'parent',
            'children',
        ):
            state.pop(attr, '')

        self.__dict__ = state
        if not hasattr(self, 'manager'):
            setattr(self, 'manager', None)

    def enabled_for(self, inpt):
        """
        Checks to see if this switch is enabled for the provided input.

        If ``compounded``, all switch conditions must be ``True`` for the switch
        to be enabled.  Otherwise, *any* condition needs to be ``True`` for the
        switch to be enabled.

        The switch state is then checked to see if it is ``GLOBAL`` or
        ``DISABLED``.  If it is not, then the switch is ``SELECTIVE`` and each
        condition is checked.

        Keyword Arguments:
        inpt -- An instance of the ``Input`` class.
        """

        signals.switch_checked.call(self)
        signal_decorated = partial(self.__signal_and_return, inpt)

        if self.state is self.states.GLOBAL:
            return signal_decorated(True)
        elif self.state is self.states.DISABLED:
            return signal_decorated(False)

        conditions_dict = ConditionsDict.from_conditions_list(self.conditions)
        conditions = conditions_dict.get_by_input(inpt)

        if conditions:
            result = self.__enabled_func(
                cond.call(inpt)
                for cond
                in conditions
                if cond.argument(inpt).applies
            )
        else:
            result = None

        return signal_decorated(result)

    def enabled_for_all(self, *inpts):
        foo = ifilter(
            lambda x: x is not None,
            (self.enabled_for(inpt) for inpt in inpts)
        )
        return self.__enabled_func(foo)

    def save(self):
        """
        Saves this switch in its manager (if present).

        Equivalent to ``self.manager.update(self)``.  If no ``manager`` is set
        for the switch, this method is a no-op.
        """
        if self.manager:
            self.manager.update(self)

    @property
    def changes(self):
        """
        A dictionary of changes to the switch since last saved.

        Switch changes are a dict in the following format::

            {
                'property_name': {'previous': value, 'current': value}
            }

        For example, if the switch name change from ``foo`` to ``bar``, the
        changes dict will be in the following structure::

            {
                'name': {'previous': 'foo', 'current': 'bar'}
            }
        """
        return dict(list(self.__changes()))

    @property
    def changed(self):
        """
        Boolean of if the switch has changed since last saved.
        """
        return bool(list(self.__changes()))

    def reset(self):
        """
        Resets switch change tracking.

        No switch properties are altered, only the tracking of what has changed
        is reset.
        """
        self.__init_vars = vars(self).copy()

    @property
    def state_string(self):
        state_vars = vars(self.states)
        rev = dict(zip(state_vars.values(), state_vars))
        return rev[self.state]

    @property
    def __enabled_func(self):
        if self.compounded:
            return all_false_if_empty
        else:
            return any

    def __changes(self):
        for key, value in self.__init_vars.items():
            if key is '_Switch__init_vars':
                continue
            elif key not in vars(self) or getattr(self, key) != value:
                yield (key, dict(previous=value, current=getattr(self, key)))

    def __signal_and_return(self, inpt, is_enabled):
        if is_enabled:
            signals.switch_active.call(self, inpt)

        return is_enabled


class Condition(object):

    """
    A Condition is the configuration of an argument, its attribute and an
    operator. It tells you if it itself is true or false given an input.

    The ``argument`` defines what this condition is checking.  Perhaps it's a
    ``User`` or ``Request`` object. The ``attribute`` name is then extracted out
    of an instance of the argument to produce a variable. That variable is then
    compared to the operator to determine if the condition applies to the input
    or not.

    For example, for the request IP address, you would define a ``Request``
    argument, that had an ``ip`` property.  A condition would then be constructed
    like so:

    from myapp.gutter import RequestArgument
    from gutter.client.models import Condition

        >> condition = Condition(argument=RequestArgument, attribute='ip', operator=some_operator)

    When the Condition is called, it is passed the input. The argument is then
    called (constructed) with input object to produce an instance.  The
    attribute is then extracted from that instance to produce the variable.
    The extracted variable is then checked against the operator.

    To put it another way, say you wanted a condition to only allow your switch
    to people between 15 and 30 years old.  To make the condition:

        1. You would create a ``UserArgument`` class that takes a user object in
           its constructor.  The class also has an ``age`` method which returns
           the user object's age.
        2. You would then create a new Condition via:
           ``Condition(argument=UserInput, attribute='age', operator=Between(15, 30))``.
        3. You then call that condition with a ``User``, and it would return
           ``True`` if the age of the user the ``UserArgument`` instance wraps
           is between 15 and 30.
    """

    def __init__(self, argument, attribute, operator, negative=False):
        self.attribute = attribute
        self.argument = argument
        self.operator = operator
        self.negative = negative

    @property
    def __is_or_is_not(self):
        return 'is not' if self.negative else 'is'

    def __repr__(self):
        argument = ".".join((self.argument.__name__, self.attribute))
        return '<Condition "%s" %s %s>' % (argument, self.__is_or_is_not, self.operator)

    def __str__(self):
        return "%s %s %s" % (self.argument_string, self.__is_or_is_not, self.operator)

    def __eq__(self, other):
        return (
            self.argument == other.argument and
            self.attribute == other.attribute and
            self.operator == other.operator and
            self.negative is other.negative
        )

    # TODO: not sure if that's ok
    __hash__ = object.__hash__

    def call(self, inpt):
        """
        Returns if the condition applies to the ``inpt``.

        If the class ``inpt`` is an instance of is not the same class as the
        condition's own ``argument``, then ``False`` is returned.  This also
        applies to the ``NONE`` input.

        Otherwise, ``argument`` is called, with ``inpt`` as the instance and
        the value is compared to the ``operator`` and the Value is returned.  If
        the condition is ``negative``, then then ``not`` the value is returned.

        Keyword Arguments:
        inpt -- An instance of the ``Input`` class.
        """
        if inpt is Manager.NONE_INPUT:
            return False

        # Call (construct) the argument with the input object
        argument_instance = self.argument(inpt)

        if not argument_instance.applies:
            return False

        application = self.__apply(argument_instance, inpt)

        if self.negative:
            application = not application

        return application

    @property
    def argument_string(self):
        parts = [self.argument.__name__, self.attribute]
        return '.'.join(map(str, parts))

    def __apply(self, argument_instance, inpt):
        variable = getattr(argument_instance, self.attribute)

        try:
            return self.operator.applies_to(variable)
        except Exception as error:
            signals.condition_apply_error.call(self, inpt, error)
            return False


class Manager(threading.local):

    """
    The Manager holds all state for Gutter.  It knows what Switches have been
    registered, and also what Input objects are currently being applied.  It
    also offers an ``active`` method to ask it if a given switch name is
    active, given its conditions and current inputs.
    """

    key_separator = DEFAULT_SEPARATOR
    namespace_separator = '.'
    default_namespace = ['default']

    #: Special singleton used to represent a "no input" which arguments can look
    #: for and ignore
    NONE_INPUT = object()

    def __init__(
        self,
        storage=None,
        autocreate=False,
        switch_class=Switch,
        inputs=None,
        namespace=None
    ):
        if storage is None:
            # todo: make a better check
            raise TypeError('storage must be a dict like value')

        if inputs is None:
            inputs = []

        if namespace is None:
            namespace = self.default_namespace
        elif isinstance(namespace, six.string_types):
            namespace = [namespace]

        self.storage = storage
        self.autocreate = autocreate
        self.inputs = inputs
        self.switch_class = switch_class
        self.namespace = namespace

    def __getstate__(self):
        inner_dict = vars(self).copy()
        inner_dict.pop('inputs', False)
        inner_dict.pop('storage', False)
        return inner_dict

    def __getitem__(self, key):
        return self.switch(key)

    def __contains__(self, key):
        return self.__namespaced(key) in self.storage

    def __delitem__(self, key):
        del self.storage[self.__namespaced(key)]

    @property
    def switches(self):
        """
        List of all switches currently registered.
        """
        results = [
            switch for name, switch in six.iteritems(self.storage)
            if name.startswith(self.__joined_namespace)
        ]

        return results

    def switch(self, name):
        """
        Returns the switch with the provided ``name``.

        If ``autocreate`` is set to ``True`` and no switch with that name
        exists, a ``DISABLED`` switch will be with that name.

        Keyword Arguments:
        name -- A name of a switch.
        """
        try:
            switch = self.storage[self.__namespaced(name)]
        except KeyError:
            if not self.autocreate:
                raise ValueError("No switch named '%s' registered in '%s'" % (name, self.namespace))

            switch = self.__create_and_register_disabled_switch(name)

        switch.manager = self
        return switch

    def get_children(self, parent):
        namespaced_parent = self.__namespaced(parent) + ':'
        return [
            self.__denamespaced(child)
            for child
            in self.storage.keys()
            if child.startswith(namespaced_parent)
        ]

    def register(self, switch, signal=signals.switch_registered):
        '''
        Register a switch and persist it to the storage.
        '''
        if not switch.name:
            raise ValueError('Switch name cannot be blank')

        switch.manager = self
        self.__persist(switch)

        signal.call(switch)

    def unregister(self, switch_or_name):
        switch = getattr(switch_or_name, 'name', switch_or_name)

        for child in self.get_children(switch):
            self.unregister(child)

        if switch in self:
            signals.switch_unregistered.call(self.switch(switch))
            del self.storage[self.__namespaced(switch)]

    def input(self, *inputs):
        self.inputs = list(inputs)

    def flush(self):
        self.inputs = []

    def active(self, name, *inputs, **kwargs):
        switch = self.switch(name)

        if not kwargs.get('exclusive', False):
            inputs = tuple(self.inputs) + inputs

        # Also check the switches against "NONE" input. This ensures there will
        # be at least one input checked.
        if not inputs:
            inputs = (self.NONE_INPUT,)

        # If necessary, the switch first consents with its parent and returns
        # false if the switch is consenting and the parent is not enabled for
        # ``inputs``.

        if (
            switch.concent
            and switch.get_parent()
            and not self.active(switch.parent, *inputs, **kwargs)
        ):
            return False

        return switch.enabled_for_all(*inputs)

    def update(self, switch):

        self.register(switch, signal=signals.switch_updated)
        switch.reset()

    def namespaced(self, namespace):
        new_namespace = []

        # Only start with the current namesapce if it's not the default
        # namespace
        if self.namespace is not self.default_namespace:
            new_namespace = list(self.namespace)

        new_namespace.append(namespace)

        return type(self)(
            storage=self.storage,
            autocreate=self.autocreate,
            inputs=self.inputs,
            switch_class=self.switch_class,
            namespace=new_namespace,
        )

    def __persist(self, switch):
        self.storage[self.__namespaced(switch.name)] = switch
        return switch

    def __create_and_register_disabled_switch(self, name):
        switch = self.switch_class(name)
        switch.state = self.switch_class.states.DISABLED
        self.register(switch)
        return switch

    def __parent_key_for(self, switch):
        # TODO: Make this a method on the switch object
        return self.name.rsplit(self.key_separator, 1)[:-1]
        # parent_parts = switch.name.split(self.key_separator)[:-1]
        # return self.key_separator.join(parent_parts)

    def __namespaced(self, name=''):
        if not self.__joined_namespace:
            return name
        else:
            return self.namespace_separator.join(
                (self.__joined_namespace, name)
            )

    def __denamespaced(self, name=''):
        prefix = self.__joined_namespace + self.namespace_separator
        if prefix and name.startswith(prefix):
            return name.replace(prefix, '', 1)
        return name

    @property
    def __joined_namespace(self):
        return self.namespace_separator.join(self.namespace)



"""
Python 3 compatibility.
"""

import six


if six.PY3:
    import unittest
    ifilter = filter
else:
    import unittest2 as unittest
    from itertools import ifilter


NoneType = type(None)


class TestCaseMixin(object):
    if six.PY3:
        def assertItemsEqual(self, actual, expected, msg=None):
            self.assertCountEqual(actual, expected, msg=None)



import random


class Base(object):

    def __init__(self, value):
        self.value = value

    def __proxy_to_value_method(method):
        def func(self, *args, **kwargs):

            if hasattr(self, 'value'):
                return getattr(self.value, method)(*args, **kwargs)
            else:
                raise NotImplementedError

        return func

    __cmp__ = __proxy_to_value_method('__cmp__')
    __hash__ = __proxy_to_value_method('__hash__')
    __nonzero__ = __proxy_to_value_method('__nonzero__')

    # PY3
    __lt__ = __proxy_to_value_method('__lt__')
    __le__ = __proxy_to_value_method('__le__')
    __gt__ = __proxy_to_value_method('__gt__')
    __ge__ = __proxy_to_value_method('__ge__')
    __eq__ = __proxy_to_value_method('__eq__')
    __ne__ = __proxy_to_value_method('__ne__')
    __bool__ = __proxy_to_value_method('__bool__')

    @staticmethod
    def to_python(value):
        return value


class Value(Base):
    pass


class Integer(Base):

    @staticmethod
    def to_python(value):
        return int(value)


class Float(Base):

    @staticmethod
    def to_python(value):
        return float(value)


class Boolean(Base):

    def __init__(self, value, hash_value=None):
        super(Boolean, self).__init__(value)
        self.hash_value = hash_value or random.getrandbits(128)

    def __hash__(self, *args, **kwargs):
        return hash(self.hash_value)

    @staticmethod
    def to_python(value):
        return bool(value)


class String(Base):

    def __cmp__(self, other):
        return cmp(self.value, other)

    def __lt__(self, other):
        return self.value < other

    def __gt__(self, other):
        return self.value > other

    def __eq__(self, other):
        return self.value == other

    def __hash__(self):
        return hash(self.value)

    def __nonzero__(self, *args, **kwargs):
        return bool(self.value)

    __bool__ = __nonzero__

    @staticmethod
    def to_python(value):
        return str(value)



from decimal import Context as decimal_Context, Decimal, DecimalException

from gutter.client.operators import Base
from gutter.client.registry import operators


class PercentRange(Base):

    name = 'percent_range'
    group = 'misc'
    preposition = 'in the percentage range of'
    arguments = ('lower_limit', 'upper_limit')

    _context = decimal_Context()

    def _modulo(self, decimal_argument):
        """
        The mod operator is prone to floating point errors, so use decimal.

        101.1 % 100
        >>> 1.0999999999999943

        decimal_context.divmod(Decimal('100.1'), 100)
        >>> (Decimal('1'), Decimal('0.1'))
        """
        _times, remainder = self._context.divmod(decimal_argument, 100)

        # match the builtin % behavior by adding the N to the result if negative
        return remainder if remainder >= 0 else remainder + 100

    def __init__(self, lower_limit, upper_limit):
        self.upper_limit = self._context.create_decimal(str(upper_limit))
        self.lower_limit = self._context.create_decimal(str(lower_limit))

    def applies_to(self, argument):
        try:
            decimal_argument = Decimal(str(argument))
        except DecimalException:
            decimal_argument = Decimal(hash(argument))

        return self.lower_limit <= self._modulo(decimal_argument) < self.upper_limit

    def __str__(self):
        return 'in %0.1f - %0.1f%% of values' % (self.lower_limit, self.upper_limit)


class Percent(PercentRange):

    name = 'percent'
    group = 'misc'
    preposition = 'within the percentage of'
    arguments = ('percentage',)

    def __init__(self, percentage):
        self.upper_limit = float(percentage)
        self.lower_limit = 0.0

    @property
    def variables(self):
        return dict(percentage=self.upper_limit)

    def __str__(self):
        return 'in %s%% of values' % self.upper_limit


operators.register(PercentRange)
operators.register(Percent)



from gutter.client.operators import Base
from gutter.client.arguments.base import Container


class Registry(object):

    def __init__(self, cls):
        self.__items = {}
        self.__cls = cls

    items = property(lambda self: self.__items)

    def __getitem__(self, key):
        try:
            return self.__items[key]
        except KeyError:
            raise KeyError("Key '%s' not registered" % key)

    def register(self, key, obj):
        try:
            if not issubclass(obj, self.__cls):
                raise ValueError("%s is not a %s" % (obj, self.__cls))
        except TypeError:
            raise ValueError("%s is not a %s" % (obj, self.__cls))

        self.__items[key] = obj


def extract_key_from_name(func):
    def helpful_register(key_or_operator=None, operator=None):
        if not operator:
            operator = key_or_operator
            key = operator.name
        else:
            key = key_or_operator

        return func(key, operator)

    return helpful_register


arguments = Registry(Container)
operators = Registry(Base)


operators.register = extract_key_from_name(operators.register)



"""
gutter.testutils
~~~~~~~~~~~~~~~~~~

:copyright: (c) 2010-2012 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
import six

from functools import wraps
from gutter.client import get_gutter_client


class SwitchContextManager(object):
    """
    Allows temporarily enabling or disabling a switch.

    Ideal for testing.

    >>> @switches(my_switch_name=True)
    >>> def foo():
    >>>     print gutter.active('my_switch_name')

    >>> def foo():
    >>>     with switches(my_switch_name=True):
    >>>         print gutter.active('my_switch_name')

    You may also optionally pass an instance of ``SwitchManager``
    as the first argument.

    >>> def foo():
    >>>     with switches(gutter, my_switch_name=True):
    >>>         print gutter.active('my_switch_name')
    """
    def __init__(self, gutter=None, **keys):
        self._gutter = gutter
        self.keys = keys

    @property
    def gutter(self):
        if self._gutter is None:
            self._gutter = get_gutter_client()
        elif isinstance(self._gutter, six.string_types):
            self._gutter = get_gutter_client(alias=self._gutter)
        return self._gutter


    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner

    def __enter__(self):

        self.previous_active_func = self.gutter.active

        def patched_active(gutter):
            real_active = gutter.active

            def wrapped(key, *args, **kwargs):
                if key in self.keys:
                    return self.keys[key]

                return real_active(key, *args, **kwargs)

            return wrapped

        self.gutter.active = patched_active(self.gutter)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.gutter.active = self.previous_active_func

switches = SwitchContextManager



import six

from gutter.client.compat import NoneType


class classproperty(object):

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class argument(object):

    def __init__(self, variable, getter):
        if issubclass(type(getter), six.string_types):
            self.getter = lambda self: getattr(self.input, getter)
        else:
            self.getter = getter

        self.owner = None
        self.variable = variable

    def __get__(self, instance, owner):
        self.owner = owner

        if instance:
            return self.variable(self.getter(instance))
        else:
            return self

    def __str__(self):
        if self.name:
            return "%s.%s" % (self.owner.__name__, self.name)
        else:
            return repr(self)

    @property
    def name(self):
        for name, attr in vars(self.owner).items():
            if attr is self:
                return name


class Container(object):

    """
    Base class for Arguments, which are responsible for understanding inputs
    and returning Argument Variables.  Argument variables are compared against
    the Operator inside of a Condition, for deciding if a condition applies to
    the specified input.
    """

    COMPATIBLE_TYPE = NoneType

    def __init__(self, inpt):
        self.input = inpt

    @classproperty
    def arguments(cls):
        return dict(
            (key, value) for key, value in vars(cls).items()
            if type(value) is argument
        )

    @property
    def applies(self):
        return isinstance(self.input, self.COMPATIBLE_TYPE)



from functools import partial

from . import variables
from .base import Container, argument  # noqa

Value = partial(argument, variables.Value)
Boolean = partial(argument, variables.Boolean)
String = partial(argument, variables.String)
Integer = partial(argument, variables.Integer)
Float = partial(argument, variables.Float)



from gutter.client import get_gutter_client

from gutter.storage.datastore_like_dict import DatastoreDict

datastore_dict = DatastoreDict(namespace='gutter_feature_flag', kind='switch')

# GUTTER
manager = get_gutter_client(
    storage=datastore_dict,
    autocreate=True
)



class OperatorInitError(ValueError):

    def __init__(self, argument):
        message = "Missing argument %s to construct operator." % argument
        super(OperatorInitError, self).__init__(message)


class Base(object):

    arguments = ()

    def __init__(self, *args, **kwargs):
        try:
            for argument in self.arguments:
                setattr(self, argument, kwargs.pop(argument))
        except KeyError:
            raise OperatorInitError(argument)

    @property
    def variables(self):
        return vars(self)

    def __eq__(self, other):
        for arg in vars(self).keys():
            if getattr(self, arg) != getattr(other, arg):
                return False

        return True



from gutter.client.operators import Base
from gutter.client.registry import operators


class Equals(Base):

    name = 'equals'
    group = 'comparable'
    preposition = 'equal to'
    arguments = ('value',)

    def applies_to(self, argument):
        return argument == self.value

    def __str__(self):
        return 'equal to "%s"' % self.value


class Between(Base):

    name = 'between'
    group = 'comparable'
    preposition = 'between'
    arguments = ('lower_limit', 'upper_limit')

    def applies_to(self, argument):
        return argument > self.lower_limit and argument < self.upper_limit

    def __str__(self):
        return 'between "%s" and "%s"' % (self.lower_limit, self.upper_limit)


class LessThan(Base):

    name = 'before'
    group = 'comparable'
    preposition = 'less than'
    arguments = ('upper_limit',)

    def applies_to(self, argument):
        return argument < self.upper_limit

    def __str__(self):
        return 'less than "%s"' % self.upper_limit


class LessThanOrEqualTo(LessThan):

    name = 'less_than_or_equal_to'
    group = 'comparable'
    preposition = 'less than or equal to'

    def applies_to(self, argument):
        return argument <= self.upper_limit

    def __str__(self):
        return 'less than or equal to "%s"' % self.upper_limit


class MoreThan(Base):

    name = 'more_than'
    group = 'comparable'
    preposition = 'more than'
    arguments = ('lower_limit',)

    def applies_to(self, argument):
        return argument > self.lower_limit

    def __str__(self):
        return 'more than "%s"' % self.lower_limit


class MoreThanOrEqualTo(MoreThan):

    name = 'more_than_or_equal_to'
    group = 'comparable'
    preposition = 'more than or equal to'

    def applies_to(self, argument):
        return argument >= self.lower_limit

    def __str__(self):
        return 'more than or equal to "%s"' % self.lower_limit


operators.register(Equals)
operators.register(Between)
operators.register(LessThan)
operators.register(LessThanOrEqualTo)
operators.register(MoreThan)
operators.register(MoreThanOrEqualTo)



class Signal(object):

    def __init__(self):
        self.__callbacks = []

    def connect(self, callback):
        if not callable(callback):
            raise ValueError("Callback argument must be callable")

        self.__callbacks.append(callback)

    def call(self, *args, **kwargs):
        for callback in self.__callbacks:
            callback(*args, **kwargs)

    def reset(self):
        self.__callbacks = []


switch_registered = Signal()
switch_unregistered = Signal()
switch_updated = Signal()
condition_apply_error = Signal()
switch_checked = Signal()
switch_active = Signal()



from gutter.client.arguments.variables import Base as VariableBase
from gutter.client.operators import Base
from gutter.client.registry import operators


class EqualsStripIgnoreCase(Base):

    name = 'strip_ignorecase_equals'
    group = 'string'
    preposition = 'strip ignore case equal to'
    arguments = ('value',)

    def applies_to(self, argument):
        if isinstance(argument, VariableBase):
            argument = str(argument.value)
        else:
            argument = str(argument)

        return argument.lower().strip() == self.value.lower().strip()

    def __str__(self):
        return '%s "%s"' % (self.preposition, self.value.lower())

operators.register(EqualsStripIgnoreCase)



#!/usr/bin/env python

import sys
import os
from setuptools import find_packages


try:
    from notsetuptools import setup
except ImportError:
    from setuptools import setup

tests_require = [
    'nose', 'exam', 'mock', 'django', 'redis'
]

if sys.version_info < (3, 3):
    tests_require.append('unittest2')
    tests_require.append('nose-performance')


setup_requires = []
if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

    os.environ.setdefault('DJANGO_SETTINGS_MODULE', __name__)
    INSTALLED_APPS = ('gutter.client',)
    SECRET_KEY = 'secret!'

if 'flake8' in sys.argv[1:]:
    setup_requires.append('flake8')
    setup_requires.append('dont-fudge-up')

setup(
    name='woven-gutter-gae',
    version='0.0.2',
    author='DISQUS, soulhave',
    author_email='opensource@disqus.com, soulhave@yahoo.com.br',
    url='https://github.com/soulhave/woven-gutter-gae',
    description='Client to gutter feature switches backend, '
                'using GCP (Datastore and appengine)',
    packages=find_packages(exclude=["tests"]),
    zip_safe=False,
    install_requires=[
        'durabledict>=0.9.0',
        'jsonpickle',
        'werkzeug',
        'six',
        'flask>=1.0.3',
        'flask-restplus>=0.12.1',
        'google-cloud-logging>=1.11.0',
        'google-cloud-datastore>=1.8.0'
    ],
    setup_requires=setup_requires,
    namespace_packages=['gutter'],
    license='Apache License 2.0',
    tests_require=tests_require,
    extras_require={'test': tests_require},
    test_suite='nose.collector',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)



"""
gutter
~~~~~~~~

:copyright: (c) 2010-2012 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

try:
    VERSION = __import__('pkg_resources').get_distribution('gutter').version
except Exception as e:
    VERSION = 'unknown'

CLIENT_CACHE = {}


def get_gutter_client(
        alias='default',
        cache=CLIENT_CACHE,
        **kwargs
):
    """
    Creates gutter clients and memoizes them in a registry for future quick access.

    Args:
        alias (str or None): Name of the client. Used for caching.
            If name is falsy then do not use the cache.
        cache (dict): cache to store gutter managers in.
        **kwargs: kwargs to be passed the Manger class.

    Returns (Manager):
        A gutter client.

    """
    from gutter.client.models import Manager

    if not alias:
        return Manager(**kwargs)
    elif alias not in cache:
        cache[alias] = Manager(**kwargs)

    return cache[alias]



# -*- coding: utf-8 -*-

from __future__ import absolute_import

# External Libraries
from durabledict.encoding import PickleEncoding
import jsonpickle as pickle


class JsonPickleEncoding(PickleEncoding):

    ENCODING = 'utf-8'

    @staticmethod
    def encode(data):
        json_string = pickle.dumps(data)
        return json_string.encode(JsonPickleEncoding.ENCODING)

    @staticmethod
    def decode(data):
        try:
            json_string = data.decode(JsonPickleEncoding.ENCODING)
            return pickle.loads(json_string)
        except Exception:
            return PickleEncoding.decode(data)


class KeyEncoderProxy:
    """
    Helps to ensure that keys are encoded and decoded.

    This is mainly useful on Python3 where we use text values as the key but on
    some storages like redis we can only store byte values. Without a
    consistent encoding switches cannot be found again, because the returned
    keys are `bytes` objects instead of `str` objects.
    """

    def __init__(self, storage, encoding='utf-8'):
        self._encoding = encoding
        self._storage = storage

    def __setitem__(self, key, val):
        key = key.encode(self._encoding)
        self._storage[key] = val

    def __delitem__(self, key):
        key = key.encode(self._encoding)
        del self._storage[key]

    def __getitem__(self, key):
        key = key.encode(self._encoding)
        return self._storage[key]

    def __contains__(self, key):
        key = key.encode(self._encoding)
        return key in self._storage

    def items(self):
        return (
            (key.decode(self._encoding), val)
            for key, val in self._storage.items())

    def keys(self):
        return (
            key.decode(self._encoding)
            for key in self._storage.keys())
